import pytest
from datetime import date
import numpy as np

import owlplanner as owl

# solver = 'MOSEK'
solver = 'HiGHS'


def test_constructor1():
    inames = ['Joe']
    yobs = [1961]
    expectancy = [80]
    name = 'test_1'
    p = owl.Plan(inames, yobs, expectancy, name)
    assert p.inames == inames
    assert np.array_equal(p.yobs, yobs)
    assert np.array_equal(p.expectancy, expectancy)
    assert p.N_i == 1
    assert p._name == name


def test_constructor1_2():
    inames = ['Joe', 'Jane']
    yobs = [1961, 1962]
    expectancy = [80, 82]
    name = 'test_2'
    p = owl.Plan(inames, yobs, expectancy, name)
    assert p.inames == inames
    assert np.array_equal(p.yobs, yobs)
    assert np.array_equal(p.expectancy, expectancy)
    assert p.N_i == 2
    assert p._name == name


def createPlan(ni, name, ny, topAge):
    # Make tests somehow year independent.
    assert ny >= 2
    thisyear = date.today().year
    ny -= 1
    if ni == 1:
        inames = ['Joe']
        expectancy = [topAge]
        yobs = [thisyear - topAge + ny]
    else:
        inames = ['Joe', 'Jane']
        # Make Jane pass 2 years before Joe.
        expectancy = [topAge - 2, topAge]
        yobs = [thisyear - topAge + ny, thisyear - topAge + ny]

    p = owl.Plan(inames, yobs, expectancy, name)
    # Use a flat profile for simplicity.
    p.setSpendingProfile('flat', 100)

    return p


def test_date_1():
    inames = ['Joe', 'Jane']
    yobs = [1961, 1962]
    expectancy = [80, 82]
    name = 'test_3'
    p = owl.Plan(inames, yobs, expectancy, name)
    assert p.inames == inames
    assert np.array_equal(p.yobs, yobs)
    assert np.array_equal(p.expectancy, expectancy)
    assert p.N_i == 2
    assert p._name == name
    assert p.startDate is None


def test_date_2():
    n = 10
    p = createPlan(1, 'date_2', n, 70)
    amount = 3.0
    p.setAccountBalances(taxable=[0], taxDeferred=[amount], taxFree=[0], startDate="1-1")
    assert p.startDate == "1-1"


def test_date_3():
    n = 10
    p = createPlan(1, 'date_3', n, 70)
    amount = 3.0
    p.setAccountBalances(taxable=[0], taxDeferred=[amount], taxFree=[0])
    assert p.startDate == date.today().strftime('%Y-%m-%d')


def test_date_4():
    n = 10
    p = createPlan(1, 'date_3', n, 70)
    amount = 3.0
    p.setAccountBalances(taxable=[0], taxDeferred=[amount], taxFree=[0], startDate="today")
    assert p.startDate == date.today().strftime('%Y-%m-%d')


def test_withdrawal1():
    n = 10
    p = createPlan(1, 'withdrawal1', n, 70)
    amount = 3.0
    p.setAccountBalances(taxable=[0], taxDeferred=[amount], taxFree=[0], startDate="1-1")
    p.setAllocationRatios('individual', generic=[[[0, 0, 0, 100], [0, 0, 0, 100]]])
    p.setRates('user', values=[0, 0, 0, 0])
    options = {'maxRothConversion': 0, 'bequest': 0, 'solver': solver, 'withSCLoop': False}
    p.solve('maxSpending', options=options)
    assert p.basis == pytest.approx(1000*amount/n, abs=0.5)
    assert p.bequest == pytest.approx(0, abs=0.5)


def test_withdrawal2():
    n = 10
    p = createPlan(1, 'withdrawal2', n, 70)
    # Small taxable income creates an income smaller than standard deduction. Testing e_n.
    amount = 40.0
    p.setAccountBalances(taxable=[0], taxDeferred=[amount], taxFree=[0], startDate="1-1")
    p.setAllocationRatios('individual', generic=[[[0, 0, 0, 100], [0, 0, 0, 100]]])
    p.setRates('user', values=[0, 0, 0, 0])
    options = {'maxRothConversion': 0, 'bequest': 0, 'solver': solver, 'withSCLoop': False}
    p.solve('maxSpending', options=options)
    assert p.basis == pytest.approx(1000*amount/n, abs=0.5)
    assert p.bequest == pytest.approx(0, abs=0.5)


def test_withdrawal2_2():
    n = 10
    p = createPlan(2, 'withdrawal2_2', n, 70)
    # Small taxable income creates an income smaller than standard deduction. Testing e_n.
    amount = 50
    p.setAccountBalances(taxable=[0, 0], taxDeferred=[amount/2, amount/2],
                         taxFree=[0, 0], startDate="1-1")
    p.setAllocationRatios('spouses', generic=[[0, 0, 0, 100], [0, 0, 0, 100]])
    p.setRates('user', values=[0, 0, 0, 0])
    options = {'maxRothConversion': 0, 'bequest': 0, 'solver': solver, 'withSCLoop': False}
    p.solve('maxSpending', options=options)
    assert p.basis == pytest.approx(1000*amount/n, abs=0.5)
    assert p.bequest == pytest.approx(0, abs=0.5)


def test_withdrawal3():
    n = 6
    p = createPlan(1, 'withdrawal3', n, 70)
    amount = 60
    p.setAccountBalances(taxable=[0], taxDeferred=[0], taxFree=[amount], startDate="1-1")
    p.setAllocationRatios('individual', generic=[[[0, 0, 0, 100], [0, 0, 0, 100]]])
    p.setRates('user', values=[0, 0, 0, 0])
    options = {'maxRothConversion': 0, 'bequest': 0, 'solver': solver, 'withSCLoop': False}
    p.solve('maxSpending', options=options)
    assert p.basis == pytest.approx(1000*amount/n, abs=0.5)
    assert p.bequest == pytest.approx(0, abs=0.5)


def test_withdrawal3_2():
    n = 6
    p = createPlan(2, 'withdrawal3', n, 70)
    amount = 60
    p.setAccountBalances(taxable=[0, 0], taxDeferred=[0, 0],
                         taxFree=[amount/2, amount/2], startDate="1-1")
    p.setAllocationRatios('spouses', generic=[[0, 0, 0, 100], [0, 0, 0, 100]])
    p.setRates('user', values=[0, 0, 0, 0])
    options = {'maxRothConversion': 0, 'bequest': 0, 'solver': solver, 'withSCLoop': False}
    p.solve('maxSpending', options=options)
    assert p.basis == pytest.approx(1000*amount/n, abs=0.5)
    assert p.bequest == pytest.approx(0, abs=0.5)


def test_taxfreegrowth1():
    n = 12
    p = createPlan(1, 'taxfreegrowth1', n, 72)
    amount = 120
    p.setAccountBalances(taxable=[0], taxDeferred=[0], taxFree=[amount], startDate="1-1")
    p.setAllocationRatios('individual', generic=[[[0, 0, 100, 0], [0, 0, 100, 0]]])
    rate = 4
    p.setRates('user', values=[0, 0, rate, 0])
    options = {'maxRothConversion': 0, 'netSpending': 0, 'solver': solver, 'withSCLoop': False}
    p.solve('maxBequest', options=options)
    assert p.basis == pytest.approx(0, abs=0.01)
    assert p.bequest == pytest.approx(1000*amount*(1+rate/100)**n, abs=0.5)


def test_taxfreegrowth1_2():
    n = 12
    p = createPlan(2, 'taxfreegrowth1', n, 72)
    amount = 120
    p.setAccountBalances(taxable=[0, 0], taxDeferred=[0, 0],
                         taxFree=[amount/2, amount/2], startDate="1-1")
    p.setAllocationRatios('spouses', generic=[[0, 0, 100, 0], [0, 0, 100, 0]])
    rate = 4
    p.setRates('user', values=[0, 0, rate, 0])
    options = {'maxRothConversion': 0, 'netSpending': 0, 'solver': solver, 'withSCLoop': False}
    p.solve('maxBequest', options=options)
    assert p.basis == pytest.approx(0, abs=0.01)
    assert p.bequest == pytest.approx(1000*amount*(1+rate/100)**n, abs=0.5)


def test_taxfreegrowth2():
    n = 15
    p = createPlan(1, 'taxfreegrowth2', n, 75)
    amount = 120
    p.setAccountBalances(taxable=[0], taxDeferred=[0], taxFree=[amount], startDate="1-1")
    p.setAllocationRatios('individual', generic=[[[0, 50, 50, 0], [0, 50, 50, 0]]])
    rate = 4
    p.setRates('user', values=[0, rate, rate, 0])
    options = {'maxRothConversion': 0, 'netSpending': 0, 'solver': solver, 'withSCLoop': False}
    p.solve('maxBequest', options=options)
    assert p.basis == pytest.approx(0, abs=0.01)
    assert p.bequest == pytest.approx(1000*amount*(1+rate/100)**n, abs=0.5)


def test_taxfreegrowth2_2():
    n = 15
    p = createPlan(2, 'taxfreegrowth2', n, 75)
    amount = 120
    p.setAccountBalances(taxable=[0, 0], taxDeferred=[0, 0],
                         taxFree=[amount/2, amount/2], startDate="1-1")
    p.setAllocationRatios('spouses', generic=[[0, 50, 50, 0], [0, 50, 50, 0]])
    rate = 4
    p.setRates('user', values=[0, rate, rate, 0])
    options = {'maxRothConversion': 0, 'netSpending': 0, 'solver': solver, 'withSCLoop': False}
    p.solve('maxBequest', options=options)
    assert p.basis == pytest.approx(0, abs=0.01)
    assert p.bequest == pytest.approx(1000*amount*(1+rate/100)**n, abs=0.5)


def test_taxfreegrowth3():
    n = 15
    p = createPlan(1, 'taxfreegrowth3', n, 75)
    amount = 120
    p.setAccountBalances(taxable=[0], taxDeferred=[0], taxFree=[amount], startDate="1-1")
    p.setAllocationRatios('individual', generic=[[[50, 50, 0, 0], [50, 50, 0, 0]]])
    rate = 4
    p.setRates('user', values=[rate, rate, 0, 0])
    options = {'maxRothConversion': 0, 'netSpending': 0, 'solver': solver, 'withSCLoop': False}
    p.solve('maxBequest', options=options)
    assert p.basis == pytest.approx(0, abs=0.01)
    assert p.bequest == pytest.approx(1000*amount*(1+rate/100)**n, abs=0.5)


def test_taxfreegrowth3_2():
    n = 15
    p = createPlan(2, 'taxfreegrowth3', n, 75)
    amount = 120
    p.setAccountBalances(taxable=[0, 0], taxDeferred=[0, 0],
                         taxFree=[amount/2, amount/2], startDate="1-1")
    p.setAllocationRatios('spouses', generic=[[50, 50, 0, 0], [50, 50, 0, 0]])
    rate = 4
    p.setRates('user', values=[rate, rate, 0, 0])
    options = {'maxRothConversion': 0, 'netSpending': 0, 'solver': solver, 'withSCLoop': False}
    p.solve('maxBequest', options=options)
    assert p.basis == pytest.approx(0, abs=0.01)
    assert p.bequest == pytest.approx(1000*amount*(1+rate/100)**n, abs=0.5)


def test_taxfreegrowth4():
    n = 16
    p = createPlan(1, 'taxfreegrowth4', n, 76)
    amount = 120
    p.setAccountBalances(taxable=[0], taxDeferred=[0], taxFree=[amount], startDate="1-1")
    p.setAllocationRatios('individual', generic=[[[0, 50, 50, 0], [0, 50, 50, 0]]])
    rate = 4
    p.setRates('user', values=[0, 0, rate, 0])
    options = {'maxRothConversion': 0, 'netSpending': 0, 'solver': solver, 'withSCLoop': False}
    p.solve('maxBequest', options=options)
    assert p.basis == pytest.approx(0, abs=0.01)
    # Only half of the amount will grow.
    assert p.bequest == pytest.approx(1000*amount*(1+rate/200)**n, abs=0.5)


def test_taxfreegrowth4_2():
    n = 16
    p = createPlan(2, 'taxfreegrowth4', n, 76)
    amount = 120
    p.setAccountBalances(taxable=[0, 0], taxDeferred=[0, 0],
                         taxFree=[amount/2, amount/2], startDate="1-1")
    p.setAllocationRatios('spouses', generic=[[0, 50, 50, 0], [0, 50, 50, 0]])
    rate = 4
    p.setRates('user', values=[0, 0, rate, 0])
    options = {'maxRothConversion': 0, 'netSpending': 0, 'solver': solver, 'withSCLoop': False}
    p.solve('maxBequest', options=options)
    assert p.basis == pytest.approx(0, abs=0.01)
    # Only half of the amount will grow.
    assert p.bequest == pytest.approx(1000*amount*(1+rate/200)**n, abs=0.5)


def test_taxfreegrowth5():
    n = 15
    p = createPlan(1, 'taxfreegrowth5', n, 76)
    amount = 120
    p.setAccountBalances(taxable=[0], taxDeferred=[0], taxFree=[amount], startDate="1-1")
    p.setAllocationRatios('individual', generic=[[[0, 0, 100, 0], [0, 0, 100, 0]]])
    # Inflation should cancel out growth, if both rates are equal.
    p.setRates('user', values=[0, 0, 4, 4])
    options = {'maxRothConversion': 0, 'netSpending': 0, 'solver': solver, 'withSCLoop': False}
    p.solve('maxBequest', options=options)
    assert p.basis == pytest.approx(0, abs=0.01)
    assert p.bequest == pytest.approx(1000*amount, abs=0.5)


def test_taxfreegrowth5_2():
    n = 15
    p = createPlan(2, 'taxfreegrowth5', n, 76)
    amount = 120
    p.setAccountBalances(taxable=[0, 0], taxDeferred=[0, 0],
                         taxFree=[amount/2, amount/2], startDate="1-1")
    p.setAllocationRatios('spouses', generic=[[0, 0, 100, 0], [0, 0, 100, 0]])
    # Inflation should cancel out growth, if both rates are equal.
    p.setRates('user', values=[0, 0, 4, 4])
    options = {'maxRothConversion': 0, 'netSpending': 0, 'solver': solver, 'withSCLoop': False}
    p.solve('maxBequest', options=options)
    assert p.basis == pytest.approx(0, abs=0.01)
    assert p.bequest == pytest.approx(1000*amount, abs=0.5)


def test_taxfreegrowth6():
    n = 15
    p = createPlan(1, 'taxfreegrowth6', n, 76)
    amount = 120
    p.setAccountBalances(taxable=[0], taxDeferred=[0], taxFree=[amount], startDate="1-1")
    p.setAllocationRatios('individual', generic=[[[0, 0, 0, 100], [0, 0, 0, 100]]])
    # Inflation should cancel out growth.
    p.setRates('user', values=[0, 0, 0, 4])
    options = {'maxRothConversion': 0, 'netSpending': 0, 'solver': solver, 'withSCLoop': False}
    p.solve('maxBequest', options=options)
    assert p.basis == pytest.approx(0, abs=0.01)
    assert p.bequest == pytest.approx(1000*amount, abs=0.5)


def test_taxfreegrowth6_2():
    n = 15
    p = createPlan(2, 'taxfreegrowth6', n, 76)
    amount = 120
    p.setAccountBalances(taxable=[0, 0], taxDeferred=[0, 0],
                         taxFree=[amount/2, amount/2], startDate="1-1")
    p.setAllocationRatios('spouses', generic=[[0, 0, 0, 100], [0, 0, 0, 100]])
    # Inflation should cancel out growth.
    p.setRates('user', values=[0, 0, 0, 4])
    options = {'maxRothConversion': 0, 'netSpending': 0, 'solver': solver, 'withSCLoop': False}
    p.solve('maxBequest', options=options)
    assert p.basis == pytest.approx(0, abs=0.01)
    assert p.bequest == pytest.approx(1000*amount, abs=0.5)


def test_taxfreegrowth7():
    n = 15
    p = createPlan(1, 'taxfreegrowth7', n, 76)
    amount = 120
    p.setAccountBalances(taxable=[0], taxDeferred=[0], taxFree=[amount], startDate="1-1")
    p.setAllocationRatios('individual', generic=[[[0, 100, 0, 0], [0, 100, 0, 0]]])
    # Inflation should cancel out growth.
    p.setRates('user', values=[0, 4, 0, 4])
    options = {'maxRothConversion': 0, 'netSpending': 0, 'solver': solver, 'withSCLoop': False}
    p.solve('maxBequest', options=options)
    assert p.basis == pytest.approx(0, abs=0.01)
    assert p.bequest == pytest.approx(1000*amount, abs=0.5)


def test_taxfreegrowth7_2():
    n = 15
    p = createPlan(2, 'taxfreegrowth7', n, 76)
    amount = 120
    p.setAccountBalances(taxable=[0, 0], taxDeferred=[0, 0],
                         taxFree=[amount/2, amount/2], startDate="1-1")
    p.setAllocationRatios('spouses', generic=[[0, 100, 0, 0], [0, 100, 0, 0]])
    # Inflation should cancel out growth.
    p.setRates('user', values=[0, 4, 0, 4])
    options = {'maxRothConversion': 0, 'netSpending': 0, 'solver': solver, 'withSCLoop': False}
    p.solve('maxBequest', options=options)
    assert p.basis == pytest.approx(0, abs=0.01)
    assert p.bequest == pytest.approx(1000*amount, abs=0.5)


def test_taxfreegrowth8():
    n = 15
    p = createPlan(1, 'taxfreegrowth8', n, 76)
    amount = 120
    p.setAccountBalances(taxable=[0], taxDeferred=[0], taxFree=[amount], startDate="1-1")
    p.setAllocationRatios('individual', generic=[[[100, 0, 0, 0], [100, 0, 0, 0]]])
    # Inflation should cancel out growth.
    p.setRates('user', values=[4, 0, 0, 4])
    options = {'maxRothConversion': 0, 'netSpending': 0, 'solver': solver, 'withSCLoop': False}
    p.solve('maxBequest', options=options)
    assert p.basis == pytest.approx(1e-7, abs=0.01)
    assert p.bequest == pytest.approx(1000*amount, abs=0.5)


def test_taxfreegrowth8_2():
    n = 15
    p = createPlan(2, 'taxfreegrowth8', n, 76)
    amount = 120
    p.setAccountBalances(taxable=[0, 0], taxDeferred=[0, 0],
                         taxFree=[amount/2, amount/2], startDate="1-1")
    p.setAllocationRatios('spouses', generic=[[100, 0, 0, 0], [100, 0, 0, 0]])
    # Inflation should cancel out growth.
    p.setRates('user', values=[4, 0, 0, 4])
    options = {'maxRothConversion': 0, 'netSpending': 0, 'solver': solver, 'withSCLoop': False}
    p.solve('maxBequest', options=options)
    assert p.basis == pytest.approx(0, abs=0.01)
    assert p.bequest == pytest.approx(1000*amount, abs=0.5)


def test_annuity1():
    n = 12
    p = createPlan(1, 'annuity1', n, 76)
    amount = 120
    p.setAccountBalances(taxable=[0], taxDeferred=[0], taxFree=[amount], startDate="1-1")
    p.setAllocationRatios('individual', generic=[[[0, 0, 100, 0], [0, 0, 100, 0]]])
    rate = 4
    p.setRates('user', values=[0, 0, rate, 0])
    options = {'maxRothConversion': 0, 'bequest': 0, 'solver': solver, 'withSCLoop': False}
    p.solve('maxSpending', options=options)
    rate /= 100
    # Annuity equation has to be modified to account that first withdrawal
    # happens at the start of first period:  PV = ((1 + i)/i)*(1 - (1 + i)^-n).
    annuity = 1000*amount*rate/((1 - (1 + rate)**-n)*(1 + rate))
    assert p.basis == pytest.approx(annuity, abs=0.5)
    assert p.bequest == pytest.approx(0, abs=0.5)


def test_annuity1_2():
    n = 12
    p = createPlan(2, 'annuity1', n, 76)
    amount = 120
    p.setAccountBalances(taxable=[0, 0], taxDeferred=[0, 0],
                         taxFree=[amount/2, amount/2], startDate="1-1")
    p.setAllocationRatios('spouses', generic=[[0, 0, 100, 0], [0, 0, 100, 0]])
    rate = 4
    p.setRates('user', values=[0, 0, rate, 0])
    options = {'maxRothConversion': 0, 'bequest': 0, 'solver': solver, 'withSCLoop': False}
    p.solve('maxSpending', options=options)
    rate /= 100
    # Annuity equation has to be modified to account that first withdrawal
    # happens at the start of first period:  PV = ((1 + i)/i)*(1 - (1 + i)^-n).
    annuity = 1000*amount*rate/((1 - (1 + rate)**-n)*(1 + rate))
    assert p.basis == pytest.approx(annuity, abs=0.5)
    assert p.bequest == pytest.approx(0, abs=0.5)


def test_annuity2():
    n = 18
    p = createPlan(1, 'annuity2', n, 76)
    amount = 120
    p.setAccountBalances(taxable=[0], taxDeferred=[0], taxFree=[amount], startDate="1-1")
    p.setAllocationRatios('individual', generic=[[[0, 0, 100, 0], [0, 0, 100, 0]]])
    rate = 4
    p.setRates('user', values=[0, 0, rate, 0])
    options = {'maxRothConversion': 0, 'bequest': 0, 'solver': solver, 'withSCLoop': False}
    p.solve('maxSpending', options=options)
    rate /= 100
    # Annuity equation has to be modified to account that first withdrawal
    # happens before the first period:  PV = ((1 + i)/i)*(1 - (1 + i)^-n).
    annuity = 1000*amount*rate/((1 - (1 + rate)**-n)*(1 + rate))
    assert p.basis == pytest.approx(annuity, abs=0.5)
    assert p.bequest == pytest.approx(0, abs=0.5)


def test_annuity2_2():
    n = 18
    p = createPlan(2, 'annuity2', n, 78)
    amount = 120
    p.setAccountBalances(taxable=[0, 0], taxDeferred=[0, 0],
                         taxFree=[amount/2, amount/2], startDate="1-1")
    p.setAllocationRatios('spouses', generic=[[0, 0, 100, 0], [0, 0, 100, 0]])
    rate = 4
    p.setRates('user', values=[0, 0, rate, 0])
    options = {'maxRothConversion': 0, 'bequest': 0, 'solver': solver, 'withSCLoop': False}
    p.solve('maxSpending', options=options)
    rate /= 100
    # Annuity equation has to be modified to account that first withdrawal
    # happens before the first period:  PV = ((1 + i)/i)*(1 - (1 + i)^-n).
    annuity = 1000*amount*rate/((1 - (1 + rate)**-n)*(1 + rate))
    assert p.basis == pytest.approx(annuity, abs=0.5)
    assert p.bequest == pytest.approx(0, abs=0.5)


def test_annuity3():
    n = 30
    p = createPlan(1, 'annuity2', n, 90)
    amount = 100
    p.setAccountBalances(taxable=[0], taxDeferred=[0], taxFree=[amount], startDate="1-1")
    p.setAllocationRatios('individual', generic=[[[0, 0, 100, 0], [0, 0, 100, 0]]])
    rate = 4
    p.setRates('user', values=[0, 0, rate, 0])
    options = {'maxRothConversion': 0, 'bequest': 0, 'solver': solver, 'withSCLoop': False}
    p.solve('maxSpending', options=options)
    rate /= 100
    # Annuity equation has to be modified to account that first withdrawal
    # happens before the first period:  PV = ((1 + i)/i)*(1 - (1 + i)^-n).
    annuity = 1000*amount*rate/((1 - (1 + rate)**-n)*(1 + rate))
    assert p.basis == pytest.approx(annuity, abs=0.5)
    assert p.bequest == pytest.approx(0, abs=0.5)


def test_annuity3_2():
    n = 30
    p = createPlan(2, 'annuity2', n, 90)
    amount = 100
    p.setAccountBalances(taxable=[0, 0], taxDeferred=[0, 0],
                         taxFree=[amount/2, amount/2], startDate="1-1")
    p.setAllocationRatios('spouses', generic=[[0, 0, 100, 0], [0, 0, 100, 0]])
    rate = 4
    p.setRates('user', values=[0, 0, rate, 0])
    options = {'maxRothConversion': 0, 'bequest': 0, 'solver': solver, 'withSCLoop': False}
    p.solve('maxSpending', options=options)
    rate /= 100
    # Annuity equation has to be modified to account that first withdrawal
    # happens before the first period:  PV = ((1 + i)/i)*(1 - (1 + i)^-n).
    annuity = 1000*amount*rate/((1 - (1 + rate)**-n)*(1 + rate))
    assert p.basis == pytest.approx(annuity, abs=0.5)
    assert p.bequest == pytest.approx(0, abs=0.5)


def test_Historical1():
    name = 'historical1'
    inames = ['Joe']
    yobs = [1962]
    expectancy = [89]
    p = owl.Plan(inames, yobs, expectancy, name)
    p.setSpendingProfile('smile', 60)

    p.setAccountBalances(taxable=[90.5], taxDeferred=[600.2],
                         taxFree=[50 + 20.6], startDate="1-1")
    p.setAllocationRatios('individual', generic=[[[60, 40, 0, 0], [70, 30, 0, 0]]])
    p.setPension([0], [65])
    options = {'maxRothConversion': 0, 'bequest': 0, 'solver': solver, 'withSCLoop': False}
    p.runHistoricalRange('maxSpending', options, 1928, 1958, figure=False)


def test_Historical2():
    name = 'historical2'
    inames = ['Jack', 'Jill']
    yobs = [1962, 1965]
    expectancy = [89, 92]
    p = owl.Plan(inames, yobs, expectancy, name)
    p.setSpendingProfile('smile', 60)
    p.setAccountBalances(taxable=[90.5, 60], taxDeferred=[600.2, 150],
                         taxFree=[50 + 20.6, 40.8], startDate="1-1")
    p.setAllocationRatios('individual',
                          generic=[[[60, 40, 0, 0], [70, 30, 0, 0]], [[60, 40, 0, 0], [70, 30, 0, 0]]],)
    p.setPension([0, 10], [65, 65])
    p.setSocialSecurity([28, 25], [70, 70])
    p.setHeirsTaxRate(33)

    options = {'maxRothConversion': 100, 'noRothConversions': 'Jill', 'withSCLoop': True}
    p.runHistoricalRange('maxSpending', options, 1928, 1958, figure=False)



"""

Owl/logging

This file contains routines for handling error messages.

Copyright &copy; 2024 - Martin-D. Lacasse

Disclaimers: This code is for educational purposes only and does not constitute financial advice.

"""

import sys


class Logger(object):
    def __init__(self, verbose=True, logstreams=None):
        self._verbose = verbose
        self._prevState = self._verbose
        if logstreams is None or logstreams == [] or len(logstreams) > 2:
            self._logstreams = [sys.stdout, sys.stderr]
            self.vprint("Using stdout and stderr as stream loggers.")
        elif len(logstreams) == 2:
            self._logstreams = logstreams
            self.vprint("Using logstreams as stream loggers.")
        elif len(logstreams) == 1:
            self._logstreams = 2 * logstreams
            self.vprint("Using logstream as stream logger.")
        else:
            raise ValueError(f"Log streams {logstreams} must be a list.")

    def setVerbose(self, verbose=True):
        """
        Set verbose to True if you want the module to be chatty,
        or False to make it silent.
        """
        self._prevState = self._verbose
        self._verbose = verbose
        self.vprint("Setting verbose to", verbose)

        return self._prevState

    def resetVerbose(self):
        """
        Reset verbose to previous state.
        """
        self._verbose = self._prevState

    def print(self, *args, **kwargs):
        """
        Unconditional printing regardless of the value of the verbose variable
        previously set.
        """
        if "file" not in kwargs:
            file = self._logstreams[0]
            kwargs["file"] = file

        print(*args, **kwargs)
        file.flush()

    def vprint(self, *args, **kwargs):
        """
        Conditional printing depending on the value of the verbose variable
        previously set.
        """
        if self._verbose:
            self.print(*args, **kwargs)

    def xprint(self, *args, **kwargs):
        """
        Print message and exit. Use to print error messages on stderr.
        The exit() used throws an exception in an interactive environment.
        """
        if "file" not in kwargs:
            file = self._logstreams[1]
            kwargs["file"] = file

        if self._verbose:
            print("ERROR:", *args, **kwargs)
            print("Exiting...")
            file.flush()

        raise Exception("Fatal error.")
        # sys.exit(-1)



"""

Owl/timelists
---

A retirement planner using linear programming optimization.

See companion document for a complete explanation and description
of all variables and parameters.

Utility functions to read and check timelists.

Copyright &copy; 2024 - Martin-D. Lacasse

Disclaimers: This code is for educational purposes only and does not constitute financial advice.

"""

from datetime import date
import pandas as pd


# Expected headers in each excel sheet, one per individual.
_timeHorizonItems = [
    "year",
    "anticipated wages",
    "taxable ctrb",
    "401k ctrb",
    "Roth 401k ctrb",
    "IRA ctrb",
    "Roth IRA ctrb",
    "Roth conv",
    "big-ticket items",
]


def read(finput, inames, horizons, mylog):
    """
    Read listed parameters from an excel spreadsheet or through
    a dictionary of dataframes through Pandas.
    Use one sheet for each individual with the following 9 columns:
    year, anticipated wages, taxable ctrb, 401k ctrb, Roth 401k ctrb,
    IRA ctrb, Roth IRA ctrb, Roth conv, and big-ticket items.
    Supports xls, xlsx, xlsm, xlsb, odf, ods, and odt file extensions.
    Return a dictionary of dataframes by individual's names.
    """

    mylog.vprint("Reading wages, contributions, conversions, and big-ticket items over time...")

    if isinstance(finput, dict):
        dfDict = finput
        finput = "dictionary of DataFrames"
        streamName = "dictionary of DataFrames"
    else:
        # Read all worksheets in memory but only process those with proper names.
        try:
            dfDict = pd.read_excel(finput, sheet_name=None, usecols=_timeHorizonItems)
        except Exception as e:
            raise Exception(f"Could not read file {finput}: {e}.") from e
        streamName = f"file '{finput}'"

    timeLists = _condition(dfDict, inames, horizons, mylog)

    mylog.vprint(f"Successfully read time horizons from {streamName}.")

    return finput, timeLists


def _condition(dfDict, inames, horizons, mylog):
    """
    Make sure that time horizons contain all years up to life expectancy,
    and that values are positive (except big-ticket items).
    """
    timeLists = {}
    thisyear = date.today().year
    for i, iname in enumerate(inames):
        endyear = thisyear + horizons[i]

        if iname not in dfDict:
            raise ValueError(f"No sheet found for {iname}.")

        df = dfDict[iname]

        df = df.loc[:, ~df.columns.str.contains("^Unnamed")]
        for col in df.columns:
            if col == "" or col not in _timeHorizonItems:
                df.drop(col, axis=1, inplace=True)

        for item in _timeHorizonItems:
            if item not in df.columns:
                raise ValueError(f"Item {item} not found for {iname}.")

        # Only consider lines in proper year range. Go back 5 years for Roth maturation.
        df = df[df["year"] >= (thisyear - 5)]
        df = df[df["year"] < endyear]
        df = df.drop_duplicates("year")
        missing = []
        for n in range(-5, horizons[i]):
            year = thisyear + n
            if not (df[df["year"] == year]).any(axis=None):
                df.loc[len(df)] = [year, 0, 0, 0, 0, 0, 0, 0, 0]
                missing.append(year)
            else:
                for item in _timeHorizonItems:
                    if item != "big-ticket items" and df[item].iloc[n] < 0:
                        raise ValueError(f"Item {item} for {iname} in year {df['year'].iloc[n]} is < 0.")

        if len(missing) > 0:
            mylog.vprint(f"Adding {len(missing)} missing years for {iname}: {missing}.")

        df.sort_values("year", inplace=True)
        # Replace empty (NaN) cells with 0 value.
        df.fillna(0, inplace=True)

        timeLists[iname] = df

        if df["year"].iloc[-1] != endyear - 1:
            raise ValueError(
                f"Time horizon for {iname} too short.\n\tIt should end in {endyear}, not {df['year'].iloc[-1]}"
            )

    return timeLists



import streamlit as st

import sskeys as kz

col1, col2, col3 = st.columns([0.69, 0.02, 0.29], gap="large")
with col3:
    st.image("http://github.com/mdlacasse/Owl/blob/main/docs/images/owl.png?raw=true")
    st.caption("Retirement planner with great wisdom")
with col1:
    st.write("# :orange[Owl Retirement Planner]\nA retirement financial exploration tool based on linear programming")
    kz.divider("orange")
    st.write("### :material/campaign: News")
    st.markdown("""
- This version includes major tax code changes resulting from the OBBBA :mega:.
To model long-term changes in federal tax calculations,
an OBBBA expiration year can be set to return to pre-TCJA rates.

- New capability :sparkles: to optimize for Medicare and IRMAA instead of using a self-consistent loop.
Use with patience :hourglass: (~5 min) as these calculations can take longer time due to the number
of binary variables involved.

- Ported code to Streamlit 1.50 which had major changes in widget identification.

Please report bugs :bug: and suggestions through the
GitHub [channel](http://github.com/mdlacasse/Owl/issues)
or directly by email [:incoming_envelope:](mailto:martin.d.lacasse@gmail.com).

Take the time to give it a :star: on GitHub if you use the code. That's all you have to give!
""")

    kz.divider("orange")
    st.write("### :material/rocket_launch: Quick Start")
    st.markdown("""
To respect your privacy, Owl does not store any information related to a case:
all is lost after a session is closed. For this reason,
two files can be used to store the specifications of a case so that it can be reproduced
at a later time:
- A *case* parameter file
specifying account balances, asset allocation, social security and pension, rates,
optimization parameters and related assumptions.
This file is in *toml* format which is editable with a simple text editor.
- A *Wages and Contributions* file containing a
time table with anticipated wages, future contributions
to savings accounts, Roth conversions and contributions for the last 5 years,
and anticipated big-ticket items, which can be either expenses or income.
This file is in Excel or LibreOffice format, and has one tab per individual in the plan.
If no file is provided, values will default to zero, but these values can be edited in the app.

With these two files, a scenario can be created and solved with only a few steps. We will use the case
of Jack and Jill provided here as an example:
1) Download these two files from the GitHub repository
 (right-click on the link and select `Save link as...`):
    - *Case* parameter file named
    [case_jack+jill.toml](https://github.com/mdlacasse/Owl/blob/main/examples/case_jack+jill.toml?raw=true)
    in editable *toml* format.
    - *Wages and Contributions* file named
    [jack+jill.xlsx](https://github.com/mdlacasse/Owl/blob/main/examples/jack+jill.xlsx?raw=true)
    in Excel format.
1) Navigate to the **Create Case** page and drag and drop the case parameter file
you just downloaded (*case_jack+jill.toml*).
Alternatively, you can directly select the case of *jack+jill* among the GitHub examples.
1) Navigate to the **Wages and Contributions** page and
drag and drop the *Wages and Contributions* file you downloaded (*jack+jill.xlsx*).
Alternatively, you can directly load the *Wages and Contributions* file from GitHub using the marked button.
1) Move to any page in the **Single Scenario** section to browse the simulation results.

Congratulations! :balloon: You just ran your first case.
:trophy: You can now explore each page and experiment with different parameters.

For creating your own cases, you can start
from scratch by selecting `New Case...` in the selection box while on the **Create Case** page,
and fill in the information needed on each page in the **Case Setup** section.
Alternatively, you can duplicate any existing case by using
the `Duplicate case` button, and then edit its values to fit your situation.

Once a case has been fully parameterized and successfully optimized,
its parameters can be saved by using the `Download case file...` button on the **Output Files** page.
Multiple cases can coexist and can be called and compared using the `Case selector` box
at the top of the page.

More information can be found on the :material/help: **[Documentation](Documentation)**
page located in the **Resources** section.
""")



"""
A simple object to display progress.

Copyright &copy; 2024 - Martin-D. Lacasse

Disclaimers: This code is for educational purposes only and does not constitute financial advice.

"""

from owlplanner import utils as u


class Progress(object):
    def __init__(self, mylog):
        self.mylog = mylog

    def start(self):
        self.mylog.print("|--- progress ---|")

    def show(self, x):
        self.mylog.print(f"\r\r{u.pc(x, f=0)}", end="")

    def finish(self):
        self.mylog.print()
